from office365.runtime.client_value import ClientValue

from sharepoint_rest_api import config


class SearchRequest(ClientValue):

    def __init__(self, query_text, selected_properties=None, refinement_filters=None, refiners=None, source_id=None,
                 row_limit=None, rows_per_page=None, start_row=None, timeout=None,
                 block_dedupe_mode=None, bypass_result_types=None):
        """
        :type query_text: str
        :type selected_properties: dict
        :type refinement_filters: dict
        :type refiners: str
        :type row_limit: int
        :type rows_per_page: int
        :type start_row: int
        :type timeout: int
        :type bypassResultTypes: bool
        :type blockDedupeMode: int
        """
        super().__init__()
        self.Querytext = query_text
        self.SelectProperties = selected_properties
        self.RefinementFilters = refinement_filters
        self.Refiners = refiners
        self.SourceId = source_id

        self.RowLimit = row_limit or config.SHAREPOINT_PAGE_SIZE
        self.RowsPerPage = rows_per_page
        self.StartRow = start_row

        self.Timeout = timeout

        self.BypassResultTypes = bypass_result_types
        self.BlockDedupeMode = block_dedupe_mode
        self.ClientType = None
        self.CollapseSpecification = None
        self.Culture = None

    @property
    def entity_type_name(self):
        return "Microsoft.Office.Server.Search.REST.SearchRequest"



import os

SHAREPOINT_CACHE_DISABLED = os.environ.get('SHAREPOINT_CACHE_DISABLED', False)
SHAREPOINT_TENANT = os.environ.get('SHAREPOINT_TENANT', 'https://unitst.sharepoint.com')
SHAREPOINT_SITE = os.environ.get('SHAREPOINT_SITE', 'GLB-DRP')
SHAREPOINT_SITE_TYPE = os.environ.get('SHAREPOINT_SITE_TYPE', 'sites')
SHAREPOINT_USERNAME = os.environ.get('SHAREPOINT_USERNAME', 'invalid_username')
SHAREPOINT_PASSWORD = os.environ.get('SHAREPOINT_PASSWORD', 'invalid_password')


SHAREPOINT_PAGE_SIZE = int(os.environ.get('SHAREPOINT_PAGE_SIZE', 25))



import collections


def to_camel(snake_str):
    return ''.join(x.title() for x in snake_str.split('_'))


def get_cache_key(args, **kwargs):
    joint_args = ''.join(args)
    order_dict = collections.OrderedDict(kwargs)
    return hash(joint_args) + hash(frozenset(order_dict.items()))


def first_upper(string):
    return string[:1].upper() + string[1:] if string else ''



from django.conf import settings
from rest_framework import serializers
from rest_framework.reverse import reverse

from sharepoint_rest_api.models import SharePointLibrary, SharePointSite, SharePointTenant


class SharePointTenantSerializer(serializers.ModelSerializer):
    class Meta:
        model = SharePointTenant
        exclude = ('username', 'password')


class SharePointSiteSerializer(serializers.ModelSerializer):
    class Meta:
        model = SharePointSite
        fields = '__all__'


class SharePointLibrarySerializer(serializers.ModelSerializer):
    site_name = serializers.ReadOnlyField(source='site.name')
    api_url = serializers.SerializerMethodField()

    def get_api_url(self, obj):
        reverse_url = reverse('sharepoint_rest_api:sharepoint-url-rest-list',
                              kwargs={'tenant': obj.site.tenant.name, 'site': obj.site.name, 'folder': obj.name})
        return settings.HOST + reverse_url

    class Meta:
        model = SharePointLibrary
        fields = ('name', 'site_name', 'active', 'library_url', 'api_url')



from django.conf import settings
from rest_framework import serializers
from rest_framework.reverse import reverse

from sharepoint_rest_api.serializers.fields import (
    CapitalizeSearchSharePointField,
    RawSearchSharePointField,
    SearchSharePointField,
    SharePointPropertyField,
    UpperSharePointPropertyField,
)


class BaseSharePointItemSerializer(serializers.Serializer):

    id = UpperSharePointPropertyField()
    guid = UpperSharePointPropertyField()
    created = SharePointPropertyField()
    modified = SharePointPropertyField()
    title = SharePointPropertyField()
    url = SharePointPropertyField()
    resource_url = serializers.ReadOnlyField()
    download_url = serializers.SerializerMethodField()
    file_leaf_ref = SharePointPropertyField()
    file_ref = SharePointPropertyField()


class SharePointSettingsSerializer(BaseSharePointItemSerializer):
    def get_download_url(self, obj):
        filename = obj.properties.get('FileLeafRef', obj.properties.get('Title', ''))
        if filename:
            k = filename.rfind(".")
            if k > 0:
                filename = filename[:k] + "." + filename[k + 1:]
            else:
                filename = f'{filename}.pdf'
        relative_url = reverse('sharepoint_rest_api:sharepoint-settings-files-download', kwargs={
            'folder': self.context['folder'],
            'filename': filename
        })
        return f'{settings.HOST}{relative_url}'


class SharePointUrlSerializer(BaseSharePointItemSerializer):
    def get_download_url(self, obj):
        filename = obj.properties.get('FileLeafRef', obj.properties.get('Title', ''))
        if filename:
            k = filename.rfind(".")
            if k > 0:
                filename = filename[:k] + "." + filename[k + 1:]
            else:
                filename = f'{filename}.pdf'
        relative_url = reverse('sharepoint_rest_api:sharepoint-url-files-download', kwargs={
            'tenant': self.context['tenant'],
            'site': self.context['site'],
            'folder': self.context['folder'],
            'filename': filename
        })
        return f'{settings.HOST}{relative_url}'


class SharePointFileSerializer(serializers.Serializer):
    name = SharePointPropertyField()
    type_name = serializers.ReadOnlyField()
    url = serializers.ReadOnlyField()
    linking_uri = SharePointPropertyField()
    server_relative_url = SharePointPropertyField()
    unique_id = SharePointPropertyField()
    title = SharePointPropertyField()
    time_created = SharePointPropertyField()
    time_last_modified = SharePointPropertyField()

    def get_download_url(self, obj):
        relative_url = reverse('sharepoint_rest_api:sharepoint-files-download', kwargs={
            'tenant': self.context['tenant'],
            'site': self.context['site'],
            'folder': self.context['folder'],
            'filename': obj.properties['Name'].split('.')[0]})
        return f'{settings.HOST}{relative_url}'


class SharePointSearchSerializer(serializers.Serializer):

    rank = SearchSharePointField()
    doc_id = SearchSharePointField()
    work_id = SearchSharePointField()
    title = SearchSharePointField()
    author = SearchSharePointField()
    size = SearchSharePointField()
    path = SearchSharePointField()
    description = SearchSharePointField()
    write = SearchSharePointField()
    last_modified_time = SearchSharePointField()
    collapsing_status = SearchSharePointField()
    hit_highlighted_summary = SearchSharePointField()
    hit_highlighted_properties = SearchSharePointField()
    file_extension = SearchSharePointField()
    content_type_id = SearchSharePointField()
    parent_link = SearchSharePointField()
    views_life_time = SearchSharePointField()
    views_recent = SearchSharePointField()
    section_names = SearchSharePointField()
    section_indexes = SearchSharePointField()
    site_logo = SearchSharePointField()
    site_description = SearchSharePointField()
    site_name = SearchSharePointField()
    is_document = SearchSharePointField()
    file_type = SearchSharePointField()
    is_container = SearchSharePointField()
    web_template = SearchSharePointField()
    secondary_file_extension = SearchSharePointField()
    unique_id = SearchSharePointField()
    prog_id = SearchSharePointField()
    linking_url = SearchSharePointField()
    site_id = SearchSharePointField()
    web_id = SearchSharePointField()
    original_path = SearchSharePointField()
    result_type_id_list = SearchSharePointField()
    result_type_id = SearchSharePointField()
    render_template_id = SearchSharePointField()
    partition_id = SearchSharePointField()
    url_zone = SearchSharePointField()
    culture = SearchSharePointField()
    geo_location_source = SearchSharePointField()

    contentclass = RawSearchSharePointField()
    deeplinks = RawSearchSharePointField()
    importance = RawSearchSharePointField()
    docaclmeta = RawSearchSharePointField()

    pictureThumbnailURL = CapitalizeSearchSharePointField()
    serverRedirectedURL = CapitalizeSearchSharePointField()
    serverRedirectedEmbedURL = CapitalizeSearchSharePointField()
    serverRedirectedPreviewURL = CapitalizeSearchSharePointField()
    sPWebUrl = CapitalizeSearchSharePointField()

    download_url = serializers.SerializerMethodField()

    def get_download_url(self, obj):
        try:
            path = [item['Value'] for item in obj if item['Key'] == 'Path'][0]
            directories = path.split('/')
            relative_url = reverse('sharepoint_rest_api:sharepoint-settings-files-download', kwargs={
                'folder': directories[-2],
                'filename': directories[-1]
            })
            return f'{settings.HOST}{relative_url}'
        except BaseException as e:
            return str(e)



from rest_framework import serializers

from sharepoint_rest_api.utils import first_upper, to_camel


class SharePointPropertyField(serializers.ReadOnlyField):
    """Get attribute from the object or properties, convering camlcase from_date --> fromDate"""

    def get_attribute(self, instance):
        camel_case = to_camel(self.source)
        if getattr(instance, 'properties') and camel_case in instance.properties:
            return instance.properties[camel_case]
        return super().get_attribute(instance)


class UpperSharePointPropertyField(serializers.ReadOnlyField):
    """Get attribute from the object or properties, it changes to upper case e.g uuid --> UUID"""

    def get_attribute(self, instance):
        upper_case = self.source.upper()
        if getattr(instance, 'properties') and upper_case in instance.properties:
            return instance.properties[upper_case]
        return super().get_attribute(instance)


class SharePointPropertyManyField(serializers.ReadOnlyField):
    """Get attribute from the object or properties, handles multivalue"""

    def get_attribute(self, instance):
        camel_case = to_camel(self.source)
        if getattr(instance, 'properties') and camel_case in instance.properties:
            values = instance.properties[camel_case]
            if values:
                values = values.replace('; ', ';').split(';')
            return values
        return super().get_attribute(instance)


class RawSearchSharePointField(serializers.ReadOnlyField):
    def get_attribute(self, instance):
        items = [item['Value'] for item in instance if item['Key'] == self.source]
        return items[0] if items else 'N/A'


class SearchSharePointField(serializers.ReadOnlyField):
    def get_attribute(self, instance):
        field_name = to_camel(self.source)
        items = [item['Value'] for item in instance if item['Key'] == field_name]
        return items[0] if items else 'N/A'


class CapitalizeSearchSharePointField(serializers.ReadOnlyField):
    def get_attribute(self, instance):
        field_name = first_upper(self.source)
        items = [item['Value'] for item in instance if item['Key'] == field_name]
        return items[0] if items else 'N/A'



from django.core.cache import caches
from rest_framework.exceptions import PermissionDenied

from sharepoint_rest_api import config
from sharepoint_rest_api.client import SharePointClient, SharePointClientException
from sharepoint_rest_api.models import SharePointLibrary
from sharepoint_rest_api.serializers.sharepoint import SharePointSettingsSerializer
from sharepoint_rest_api.views.base import (
    AbstractSharePointViewSet,
    CamlQuerySharePointViewSet,
    FileSharePointMixin,
    RestQuerySharePointViewSet,
    SharePointSearchViewSet,
)

cache = caches['default']


class SettingsBasedSharePointViewSet(AbstractSharePointViewSet):

    def is_public(self):
        return SharePointLibrary.objects.filter(name=self.folder, public=True)

    @property
    def tenant(self):
        return config.SHAREPOINT_TENANT

    @property
    def site(self):
        return config.SHAREPOINT_SITE

    @property
    def folder(self):
        return self.kwargs.get('folder', 'Documents')

    @property
    def site_type(self):
        return config.SHAREPOINT_SITE_TYPE

    @property
    def client(self):
        key = self.get_cache_key(**{'client': 'client'})
        client = cache.get(key)
        if client is None:
            dl_info = {
                'url': f'{self.tenant}/{self.site_type}/{self.site}',
                'relative_url': f'{self.site_type}/{self.site}',
                'folder': self.folder
            }
            try:
                client = SharePointClient(**dl_info)
                if config.SHAREPOINT_CACHE_DISABLED:
                    cache.set(key, client)
            except SharePointClientException:
                raise PermissionDenied

        return client


class SharePointSettingsRestViewSet(SettingsBasedSharePointViewSet, RestQuerySharePointViewSet):
    serializer_class = SharePointSettingsSerializer


class SharePointSettingsCamlViewSet(SettingsBasedSharePointViewSet, CamlQuerySharePointViewSet):
    serializer_class = SharePointSettingsSerializer


class SharePointSettingsFileViewSet(FileSharePointMixin, SettingsBasedSharePointViewSet):
    pass


class SharePointSettingsSearchViewSet(SharePointSearchViewSet, SettingsBasedSharePointViewSet):
    pass



from django.core.cache import caches
from rest_framework.exceptions import PermissionDenied

from sharepoint_rest_api import config
from sharepoint_rest_api.client import SharePointClient, SharePointClientException
from sharepoint_rest_api.models import SharePointLibrary
from sharepoint_rest_api.serializers.sharepoint import SharePointUrlSerializer
from sharepoint_rest_api.views.base import (
    AbstractSharePointViewSet,
    CamlQuerySharePointViewSet,
    FileSharePointMixin,
    RestQuerySharePointViewSet,
    SharePointSearchViewSet,
)

cache = caches['default']


class UrlBasedSharePointViewSet(AbstractSharePointViewSet):

    def get_library(self):
        return SharePointLibrary.objects.get(
            site__tenant__url__contains=self.tenant, site__name=self.site, name=self.folder)

    def is_public(self):
        return self.get_library().public

    @property
    def tenant(self):
        return self.kwargs.get('tenant')

    @property
    def site(self):
        return self.kwargs.get('site')

    @property
    def folder(self):
        return self.kwargs.get('folder')

    @property
    def client(self):
        key = self.get_cache_key(**{'client': 'client'})
        client = cache.get(key)
        if client is None:
            dl = self.get_library()
            dl_info = {
                'url': dl.site.site_url(),
                'relative_url': dl.site.relative_url(),
                'folder': dl.name
            }
            if dl.site.tenant.username:
                dl_info['username'] = dl.site.tenant.username
                dl_info['password'] = dl.site.tenant.password
            try:
                client = SharePointClient(**dl_info)
                if config.SHAREPOINT_CACHE_DISABLED:
                    cache.set(key, client)
            except SharePointClientException:
                raise PermissionDenied

        return client


class SharePointUrlRestViewSet(RestQuerySharePointViewSet, UrlBasedSharePointViewSet):
    serializer_class = SharePointUrlSerializer


class SharePointUrlCamlViewSet(CamlQuerySharePointViewSet, UrlBasedSharePointViewSet):
    serializer_class = SharePointUrlSerializer


class SharePointUrlFileViewSet(FileSharePointMixin, UrlBasedSharePointViewSet):
    pass


class SharePointUrlSearchViewSet(SharePointSearchViewSet, UrlBasedSharePointViewSet):
    pass



import logging
import os

from office365.runtime.auth.user_credential import UserCredential
from office365.sharepoint.client_context import ClientContext
from office365.sharepoint.files.file import File
from office365.sharepoint.files.file_creation_information import FileCreationInformation
from office365.sharepoint.search.searchService import SearchService

from sharepoint_rest_api import config
from sharepoint_rest_api.builders.camlquery_builder import CamlQueryBuilder
from sharepoint_rest_api.builders.querystring_builder import QueryStringBuilder
from sharepoint_rest_api.builders.search_request_builder import SearchRequestBuilder
from sharepoint_rest_api.config import SHAREPOINT_PAGE_SIZE

logger = logging.getLogger(__name__)


default_select = ['*', 'FileLeafRef']


class SharePointClientException(BaseException):
    """SharePoint Exception when initializing the client"""


class SharePointClient:
    """Client to access SharePoint Document Library"""

    def __init__(self, *args, **kwargs) -> None:
        self.relative_url = kwargs.get('relative_url', None)
        self.site_path = kwargs.get('url', config.SHAREPOINT_TENANT)
        username = kwargs.get('username', config.SHAREPOINT_USERNAME)
        password = kwargs.get('password', config.SHAREPOINT_PASSWORD)
        self.folder = kwargs.get('folder', 'Documents')
        self.context = ClientContext.connect_with_credentials(self.site_path, UserCredential(username, password))

    def __reduce__(self):
        return (SharePointClient, (self.relative_url, self.site_path, self.folder, ))

    def get_folder(self, list_title):
        list_obj = self.context.web.lists.get_by_title(list_title)
        folder = list_obj.rootFolder
        self.context.load(folder)
        self.context.execute_query()
        logger.info(f'List url: {folder.properties["ServerRelativeUrl"]}')
        return folder

    def read_folders(self, list_title):
        self.get_folder(list_title)
        folders = self.context.web.folders
        self.context.load(folders)
        self.context.execute_query()
        for folder in folders:
            logger.info(f'Folder name: {folder.properties["Name"]}')

        return folders

    def read_files(self, filters=None):
        filters = filters or dict()
        querystring = QueryStringBuilder(filters).get_querystring()
        folder = self.get_folder(self.folder)
        files = folder.files.filter(querystring)
        self.context.load(files)
        self.context.execute_query()
        for cur_file in files:
            logger.info(f'File name: {cur_file.properties["Name"]}')

        return files

    def read_items(self, filters=None, select=None):
        filters = filters or dict()
        querystring = QueryStringBuilder(filters).get_querystring()
        list_object = self.context.web.lists.get_by_title(self.folder)
        select_string = select or default_select
        items = list_object.get_items().filter(querystring).select(select_string)
        self.context.load(items)
        self.context.execute_query()
        return items

    def read_file(self, filename):
        folder = self.get_folder(self.folder)
        cur_file = folder.files.get_by_url(f'/{self.relative_url}/{self.folder}/{filename}')
        self.context.load(cur_file)
        self.context.execute_query()
        logger.info(f'File name: {cur_file.properties["Name"]}')
        return cur_file

    def read_caml_items(self, filters=None, scope=None):
        filters = filters or dict()
        list_obj = self.context.web.lists.get_by_title(self.folder)
        qry = CamlQueryBuilder(filters, scope).get_query()
        items = list_obj.get_items(qry)
        self.context.execute_query()
        return items

    def search(self, filters=None, select=None, source_id=None, page=1):
        filters = filters or dict()
        search = SearchService(self.context)
        request = SearchRequestBuilder(filters, select, source_id, (page - 1) * SHAREPOINT_PAGE_SIZE).build()
        result = search.post_query(request)
        self.context.execute_query()
        relevant_results = result.PrimaryQueryResult.RelevantResults
        results = relevant_results['Table']['Rows'].values()
        logger.info(f'Retrieved: {relevant_results["TotalRows"]} results')
        items = [list(item['Cells'].values()) for item in results]
        return items

    def upload_file_alt(self, target_folder, name, content):
        context = target_folder.context
        info = FileCreationInformation()
        info.content = content
        info.url = name
        info.overwrite = True
        target_file = target_folder.files.add(info)
        context.execute_query()
        return target_file

    def upload_file(self, path, list_title='Documents', upload_into_library=True):
        with open(path, 'rb') as content_file:
            file_content = content_file.read()

        if upload_into_library:
            target_folder = self.context.web.lists.get_by_title(list_title).rootFolder
            file = self.upload_file_alt(target_folder, os.path.basename(path), file_content)
            logger.info('File url: {}'.format(file.properties['ServerRelativeUrl']))
        else:
            target_url = f'/{self.folder}/{os.path.basename(path)}'
            File.save_binary(self.context, target_url, str(file_content))

    def download_file(self, filename):
        response = File.open_binary(self.context, f'/{self.folder}/{filename}')
        with open(f'./data/{filename}', 'wb') as local_file:
            local_file.write(response.content)



import logging

from office365.sharepoint.listitems.caml.camlQuery import CamlQuery

from sharepoint_rest_api.utils import to_camel

logger = logging.getLogger(__name__)


def recursive_builder(queries, operator='And'):
    if queries:
        query = queries.pop()
        if len(queries) == 0:
            return query
        elif len(queries) == 1:
            last_query = queries.pop()
            return f'<{operator}>' + query + last_query + f'</{operator}>'
        else:
            return f'<{operator}>' + query + recursive_builder(queries, operator) + f'</{operator}>'
    return ''


class CamlQueryBuilder:
    date_operators = ['Geq', 'Gt', 'Leq', 'Lt']
    mapping_operator = {
        'gte': 'Geq',
        'gt': 'Gt',
        'lte': 'Leq',
        'lt': 'Lt',
        'not': 'Neq',
        'contains': 'Contains',
        'eq': 'Eq',
    }

    filters = {}

    def __init__(self, filters, scope=None):
        super().__init__()
        self.scope = scope

        if filters:
            self.filters = filters

    def create_query(self):
        where_condition = ''

        if self.filters.keys():
            filter_queries = []
            for filter_name, filter_value in self.filters.items():
                querystring_operator = filter_name.split('__')[-1]
                operator = self.mapping_operator.get(querystring_operator, 'Eq')

                filter_name = to_camel(filter_name.split('__')[0])
                if operator in self.date_operators:
                    column_type, value = 'DateTime', f"{filter_value}T00:00:00Z"  # 2016-03-26
                    query = '<{}><FieldRef Name="{}" /><Value Type="{}">{}</Value></{}>'.format(
                        operator, filter_name, column_type, value, operator)
                elif operator == 'Contains':
                    column_type = 'Text'
                    query = '<{}><FieldRef Name="{}" /><Value Type="{}">{}</Value></{}>'.format(
                            operator, filter_name, column_type, filter_value, operator)
                else:
                    column_type, values = 'Text', filter_value.split(',')
                    queries = ['<{}><FieldRef Name="{}" /><Value Type="{}">{}</Value></{}>'.format(
                        operator, filter_name, column_type, value, operator) for value in values]
                    query = recursive_builder(queries, 'Or')
                filter_queries.append(query)
            where_condition = recursive_builder(filter_queries)
            if len(filter_queries) > 1:
                where_condition = f"<And>{where_condition}</And>"

        scope = f' Scope="{self.scope}"' if self.scope else ''
        query = f'<View{scope}><Query><Where>{where_condition}</Where></Query></View>'
        return query

    def get_query(self):
        return CamlQuery.create_custom_query(self.create_query())



import logging

from sharepoint_rest_api.utils import to_camel

logger = logging.getLogger(__name__)


class QueryStringBuilder:
    """class to map web-querystring dictionary to sharepoint-querystring"""
    date_operators = ['ge', 'gt', 'le', 'lt']
    mapping_operator = {
        'gte': 'ge',
        'gt': 'gt',
        'lte': 'le',
        'lt': 'lt',
        'not': 'ne',
        'contains': 'substringof'
    }
    search = []
    filters = {}

    def __init__(self, filters):
        super().__init__()
        if filters:
            self.filters = filters

    def get_filter_querystring(self):
        filter_queries = []
        for filter_name, filter_value in self.filters.items():
            # operator
            querystring_operator = filter_name.split('__')[-1]
            operator = self.mapping_operator.get(querystring_operator, 'eq')
            # filter
            filter_name = to_camel(filter_name.split('__')[0])
            if operator in self.date_operators:
                values = [f"{filter_value}T00:00:00Z"]  # 2016-03-26
                query = ' or '.join([f"{filter_name} {operator} datetime'{value}'" for value in values])
            elif operator == 'substringof':
                values = filter_value.split(',')
                query = ' or '.join([f"{operator}('{value}', {filter_name})" for value in values])

            else:
                values = filter_value.split(',')
                query = ' or '.join([f"{filter_name} {operator} '{value}'" for value in values])
                if len(values) > 1:
                    query = f'({query})'
            filter_queries.append(query)
            logger.info(query)
        return str(" and ".join(filter_queries))

    def get_querystring(self):
        return self.get_filter_querystring() or ''



from django.core.cache import caches
from django.http import Http404, HttpResponse
from django_filters.rest_framework import DjangoFilterBackend
from office365.runtime.client_request_exception import ClientRequestException
from office365.sharepoint.files.file import File
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import ErrorDetail
from rest_framework.filters import OrderingFilter, SearchFilter

from sharepoint_rest_api import config
from sharepoint_rest_api.serializers.sharepoint import SharePointFileSerializer, SharePointSearchSerializer
from sharepoint_rest_api.utils import get_cache_key

cache = caches['default']


class AbstractSharePointViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = None
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)
    tenant = None
    site = None
    folder = None

    def get_serializer_context(self):
        ctx = super().get_serializer_context()
        ctx.update({
            'tenant': self.tenant,
            'site': self.site,
            'folder': self.folder
        })
        return ctx

    def handle_exception(self, exc):
        response = super().handle_exception(exc)
        if isinstance(exc, Http404):
            response.data['detail'] = ErrorDetail('No document found using selected filters', 'not_found')
        return response

    def get_cache_key(self, **kwargs):
        keys = [key for key in [self.tenant, self.site, self.folder] if key]
        key = get_cache_key(keys, **kwargs)
        return key


class CamlQuerySharePointViewSet(AbstractSharePointViewSet):

    def get_queryset(self):
        kwargs = self.request.query_params.dict()
        cache_dict = kwargs.copy()
        cache_dict['caml'] = 'true'
        try:
            key = self.get_cache_key(**cache_dict)
            response = cache.get(key)
            if response is None:
                response = self.client.read_caml_items(filters=kwargs)
                if config.SHAREPOINT_CACHE_DISABLED:
                    cache.set(key, response)
            return response
        except ClientRequestException:
            raise Http404


class RestQuerySharePointViewSet(AbstractSharePointViewSet):

    def get_queryset(self):
        kwargs = self.request.query_params.dict()
        try:
            key = self.get_cache_key(**kwargs)
            response = cache.get(key)
            if response is None:
                response = self.client.read_items(filters=kwargs)
                if config.SHAREPOINT_CACHE_DISABLED:
                    cache.set(key, response)
            return response
        except ClientRequestException:
            raise Http404


class FileSharePointMixin:
    serializer_class = SharePointFileSerializer
    lookup_field = 'filename'
    lookup_value_regex = '[^/]+'

    def get_object(self):
        filename = self.kwargs.get('filename', None)
        try:
            doc_file = self.client.read_file(f'{filename}')
        except ClientRequestException:
            raise Http404
        return doc_file

    def get_queryset(self):
        kwargs = self.request.query_params.dict()
        try:
            return self.client.read_files(filters=kwargs)
        except ClientRequestException:
            raise Http404

    @action(detail=True, methods=['get'])
    def download(self, request, *args, **kwargs):
        sh_file = self.get_object()
        relative_url = sh_file.properties['ServerRelativeUrl']
        response = File.open_binary(self.client.context, relative_url)

        django_response = HttpResponse(
            content=response.content,
            status=response.status_code,
            content_type=response.headers['Content-Type'],
        )
        django_response['Content-Disposition'] = 'attachment; filename=%s' % sh_file.properties['Name']
        return django_response


class SharePointSearchViewSet(AbstractSharePointViewSet):
    serializer_class = SharePointSearchSerializer
    selected_fields = None

    def get_filters(self, kwargs):
        return kwargs

    def get_selected(self, selected):
        return selected.split(',') if selected else self.serializer_class._declared_fields.keys()

    def get_queryset(self):
        kwargs = self.request.query_params.dict()
        selected = self.get_selected(kwargs.pop('selected', None))
        source_id = kwargs.pop('source_id', None)
        page = int(kwargs.pop('page', 1))
        filters = self.get_filters(kwargs)
        try:
            key = self.get_cache_key(**kwargs)
            response = cache.get(key)
            if response is None:
                response = self.client.search(
                    filters=filters,
                    select=selected,
                    source_id=source_id,
                    page=page
                )
                if config.SHAREPOINT_CACHE_DISABLED:
                    cache.set(key, response)
            return response
        except ClientRequestException:
            raise Http404



from rest_framework import routers

from sharepoint_rest_api.views.model_views import (
    SharePointLibraryViewSet,
    SharePointSiteViewSet,
    SharePointTenantViewSet,
)
from sharepoint_rest_api.views.settings_based import (
    SharePointSettingsCamlViewSet,
    SharePointSettingsFileViewSet,
    SharePointSettingsRestViewSet,
    SharePointSettingsSearchViewSet,
)
from sharepoint_rest_api.views.url_based import (
    SharePointUrlCamlViewSet,
    SharePointUrlFileViewSet,
    SharePointUrlRestViewSet,
)

app_name = 'sharepoint_rest_api'

router = routers.DefaultRouter()

# model
router.register(r'tenants', SharePointTenantViewSet, basename='sharepoint-tenant')
router.register(r'sites', SharePointSiteViewSet, basename='sharepoint-site')
router.register(r'libraries', SharePointLibraryViewSet, basename='sharepoint-library')

# url based
router.register(r'sharepoint/(?P<tenant>[\w\-]+)/(?P<site>[\w\-]+)/(?P<folder>[\w\W]+)/files',
                SharePointUrlFileViewSet, basename='sharepoint-url-files')
router.register(r'sharepoint/(?P<tenant>[\w\-]+)/(?P<site>[\w\-]+)/(?P<folder>[\w\W]+)/rest',
                SharePointUrlRestViewSet, basename='sharepoint-url-rest')
router.register(r'sharepoint/(?P<tenant>[\w\-]+)/(?P<site>[\w\-]+)/(?P<folder>[\w\W]+)/caml',
                SharePointUrlCamlViewSet, basename='sharepoint-url-caml')
router.register(r'sharepoint/(?P<tenant>[\w\-]+)/(?P<site>[\w\-]+)/search',
                SharePointSettingsSearchViewSet, basename='sharepoint-url-search')

# settings based
router.register(r'sharepoint/(?P<folder>[\w\W]+)/rest',
                SharePointSettingsRestViewSet, basename='sharepoint-settings-rest')
router.register(r'sharepoint/(?P<folder>[\w\W]+)/caml',
                SharePointSettingsCamlViewSet, basename='sharepoint-settings-caml')
router.register(r'sharepoint/(?P<folder>[\w\W]+)/files',
                SharePointSettingsFileViewSet, basename='sharepoint-settings-files')
router.register(r'sharepoint/search', SharePointSettingsSearchViewSet, basename='sharepoint-settings-search')

urlpatterns = router.urls



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter, SearchFilter

from sharepoint_rest_api.filters import SharePointLibraryFilter
from sharepoint_rest_api.models import SharePointLibrary, SharePointSite, SharePointTenant
from sharepoint_rest_api.serializers.model_serializers import (
    SharePointLibrarySerializer,
    SharePointSiteSerializer,
    SharePointTenantSerializer,
)


class SharePointTenantViewSet(viewsets.ReadOnlyModelViewSet):
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)
    queryset = SharePointTenant.objects.all()
    serializer_class = SharePointTenantSerializer
    search_fields = ('url', )


class SharePointSiteViewSet(viewsets.ReadOnlyModelViewSet):
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)
    queryset = SharePointSite.objects.all()
    serializer_class = SharePointSiteSerializer
    search_fields = ('name', )


class SharePointLibraryViewSet(viewsets.ReadOnlyModelViewSet):
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)
    queryset = SharePointLibrary.objects.all()
    serializer_class = SharePointLibrarySerializer
    search_fields = ('name', )
    filterset_class = SharePointLibraryFilter



from sharepoint_rest_api.libs.search_request import SearchRequest


class SearchRequestBuilder:
    filters = {}
    select = None

    mapping_operator = {
        'gte': '>=',
        'gt': '>',
        'lte': '<=',
        'lt': '<',
        'not': '<>',
        'eq': ':',
        'between': '..',
        'contains': '*'
    }

    def __init__(self, filters=None, select=None, source_id=None, start_row=None):
        self.filters = filters
        self.select = select
        self.source_id = source_id
        self.start_row = start_row

    def get_select_properties(self):
        if self.select:
            return {'results': self.select}

    def get_query(self):
        filter_queries = []
        if self.filters.keys():
            filter_queries = []
            for qs_filter_name, filter_value in self.filters.items():
                filter_name = qs_filter_name.split('__')[0]
                querystring_operator = qs_filter_name.split('__')[-1]
                operator = self.mapping_operator.get(querystring_operator, ':')
                if operator == '..':
                    filter_value_from, filter_value_to = filter_value.split('__')
                    query = '{}:{}{}{}'.format(filter_name, filter_value_from, operator, filter_value_to)
                elif operator == "*":
                    query = '{}:"{}{}"'.format(filter_name, filter_value, operator)
                else:
                    values = filter_value.split(',')
                    if len(values) == 1:
                        filter_values = f'\"{values[0]}\"'
                    else:
                        filter_values = "(" + " OR ".join(['\"{}\"'.format(value) for value in values]) + ')'
                    query = '{}{}{}'.format(filter_name, operator, filter_values)
                filter_queries.append(query)
        if not filter_queries:
            return '*'
        qry = ' AND '.join('{}'.format(query) for query in filter_queries)
        return qry

    def build(self):
        qry = self.get_query()
        selected_properties = self.get_select_properties()
        return SearchRequest(
            qry,
            selected_properties=selected_properties,
            source_id=self.source_id,
            start_row=self.start_row
        )
